@@ -13,28 +13,28 @@ from threading import Thread
 from typing import NamedTuple, Optional, Dict, List, Callable
 
 import cortex
-import requests
 import yaml
 from cortex.binary import get_cli_path
 from psycopg2._psycopg import DatabaseError
 from psycopg2.extras import NamedTupleCursor
-from psycopg2.pool import ThreadedConnectionPool
-from requests.adapters import HTTPAdapter, Retry
+from psycopg2.pool import ThreadedConnectionPool, PoolError
 
 from cortex_serving_client.command_line_wrapper import _verbose_command_wrapper
 from cortex_serving_client.deployment_failed import DeploymentFailed, COMPUTE_UNAVAILABLE_FAIL_TYPE, \
     DEPLOYMENT_TIMEOUT_FAIL_TYPE, DEPLOYMENT_ERROR_FAIL_TYPE, DEPLOYMENT_JOB_NOT_DEPLOYED_FAIL_TYPE
 from cortex_serving_client.printable_chars import remove_non_printable
-
+from cortex_serving_client.retry_utils import create_always_retry_session
 
 KIND_REALTIME_API = 'RealtimeAPI'
 KIND_BATCH_API = 'BatchAPI'
+KIND_ASYNC_API = 'AsyncAPI'
+KIND_TASK_API = 'TaskAPI'
 KIND_DEFAULT = KIND_REALTIME_API
 
 NOT_DEPLOYED = "not deployed"
 
 """
-Source: https://github.com/cortexlabs/cortex/blob/0.23/pkg/types/status/code.go
+Source: https://github.com/cortexlabs/cortex/blob/0.31/pkg/types/status/code.go
 On some places custom "not_deployed" status may be used.
 """
 
@@ -49,6 +49,8 @@ JOB_STATUS_ENQUEUING = 'status_enqueuing'
 JOB_STATUS_RUNNING = 'status_running'
 JOB_STATUS_UNEXPECTED_ERROR = 'status_unexpected_error'
 JOB_STATUS_ENQUEUED_FAILED = 'status_enqueue_failed'
+JOB_STATUS_COMPLETED_WITH_FAILURES = 'status_completed_with_failures'
+JOB_FINISHED_OK_STATUSES = (JOB_STATUS_SUCCEEDED, JOB_STATUS_COMPLETED_WITH_FAILURES)
 
 
 CORTEX_DELETE_TIMEOUT_SEC = 10 * 60
@@ -60,6 +62,8 @@ CORTEX_DEPLOY_RETRY_BASE_SLEEP_SEC = 5 * 60
 CORTEX_STATUS_CHECK_SLEEP_SEC = 15
 INFINITE_TIMEOUT_SEC = 30 * 365 * 24 * 60 * 60  # 30 years
 WAIT_BEFORE_JOB_GET = int(os.environ.get('CORTEX_WAIT_BEFORE_JOB_GET', str(30)))
+DB_TRIES = 10
+DB_RETRY_SEC = 5
 
 CORTEX_PATH = get_cli_path()
 
@@ -119,7 +123,7 @@ class CortexClient:
         if "kind" not in deployment:
             deployment["kind"] = KIND_DEFAULT
 
-        predictor_yaml_str = yaml.dump([deployment], default_flow_style=False)
+        predictor_yaml_str = yaml.dump([deployment], default_style='"')
 
         if api_timeout_sec < CORTEX_MIN_API_TIMEOUT_SEC:
             logger.info(f'API timeout {api_timeout_sec} is be smaller than minimal API timeout {CORTEX_MIN_API_TIMEOUT_SEC}. Setting it to {CORTEX_MIN_API_TIMEOUT_SEC} to avoid excessive GC.')
@@ -248,56 +252,53 @@ class CortexClient:
         ) as get_result:
 
             logger.info(f'BatchAPI {deployment["name"]} deployed. Starting a job.')
-            s = requests.Session()
-            http_adapter = HTTPAdapter(max_retries=Retry(total=5, backoff_factor=3))
-            s.mount('http://', http_adapter)
-            s.mount('https://', http_adapter)
-            job_id = None
-            for retry_job_get in range(5):
-                for retry_job_submit in range(3):
-                    try:
-                        job_json = s.post(get_result.endpoint, json=job_spec, timeout=10 * 60).json()
+            with create_always_retry_session() as session:
+                job_id = None
+                for retry_job_get in range(5):
+                    for retry_job_submit in range(3):
+                        try:
+                            job_json = session.post(get_result.endpoint, json=job_spec, timeout=10 * 60).json()
 
-                    except JSONDecodeError as e:
-                        logger.warning(f'Job submission response could not be decoded: {e}.')
-                        job_json = None
+                        except JSONDecodeError as e:
+                            logger.warning(f'Job submission response could not be decoded: {e}.')
+                            job_json = None
 
-                    if job_json is not None and 'job_id' in job_json:
-                        job_id = job_json['job_id']
-                        break
+                        if job_json is not None and 'job_id' in job_json:
+                            job_id = job_json['job_id']
+                            break
 
-                    else:
-                        sleep_time = 3 * 2 ** retry_job_submit
-                        logger.info(f'Retrying job creation request after {sleep_time}.')
-                        time.sleep(sleep_time)
+                        else:
+                            sleep_time = 3 * 2 ** retry_job_submit
+                            logger.info(f'Retrying job creation request after {sleep_time}.')
+                            time.sleep(sleep_time)
+                            continue
+
+                    if job_id is None:
+                        raise ValueError(f'job_id not in job_json {json.dumps(job_json)}')
+
+                    time.sleep(WAIT_BEFORE_JOB_GET * 2 ** retry_job_get)  # Don't call job too early: https://gitter.im/cortexlabs/cortex?at=5f7fe4c01cbba72b63cb745f
+
+                    if print_logs:
+                        self._cortex_logs_print_async(deployment['name'], job_id)
+
+                    job_status = JOB_STATUS_ENQUEUING
+                    while job_status in (JOB_STATUS_ENQUEUING, JOB_STATUS_RUNNING):
+                        job_result = self.get(deployment["name"], job_id)
+                        job_status = job_result.status
+                        time.sleep(30)
+
+                    if job_status in (NOT_DEPLOYED_STATUS, JOB_STATUS_UNEXPECTED_ERROR, JOB_STATUS_ENQUEUED_FAILED):
+                        # TODO request fixes improvements https://gitter.im/cortexlabs/cortex?at=5f7fe4c01cbba72b63cb745f
+                        # Sleep in after job submission.
+                        logger.warning(f'Job unexpectedly undeployed or failed with status {job_status}. Retrying with sleep.')
                         continue
 
-                if job_id is None:
-                    raise ValueError(f'job_id not in job_json {json.dumps(job_json)}')
+                    else:
+                        logger.info(f'BatchAPI {deployment["name"]} job {job_id} ended with status {job_status}. Deleting the BatchAPI.')
+                        return job_result
 
-                time.sleep(WAIT_BEFORE_JOB_GET * 2 ** retry_job_get)  # Don't call job too early: https://gitter.im/cortexlabs/cortex?at=5f7fe4c01cbba72b63cb745f
-
-                if print_logs:
-                    self._cortex_logs_print_async(deployment['name'], job_id)
-
-                job_status = JOB_STATUS_ENQUEUING
-                while job_status in (JOB_STATUS_ENQUEUING, JOB_STATUS_RUNNING):
-                    job_result = self.get(deployment["name"], job_id)
-                    job_status = job_result.status
-                    time.sleep(30)
-
-                if job_status in (NOT_DEPLOYED_STATUS, JOB_STATUS_UNEXPECTED_ERROR, JOB_STATUS_ENQUEUED_FAILED):
-                    # TODO request fixes improvements https://gitter.im/cortexlabs/cortex?at=5f7fe4c01cbba72b63cb745f
-                    # Sleep in after job submission.
-                    logger.warning(f'Job unexpectedly undeployed or failed with status {job_status}. Retrying with sleep.')
-                    continue
-
-                else:
-                    logger.info(f'BatchAPI {deployment["name"]} job {job_id} ended with status {job_status}. Deleting the BatchAPI.')
-                    return job_result
-
-            raise DeploymentFailed(f'Job unexpectedly undeployed or failed with status {job_status}.', DEPLOYMENT_JOB_NOT_DEPLOYED_FAIL_TYPE,
-                                   deployment['name'], -1)
+                raise DeploymentFailed(f'Job unexpectedly undeployed or failed with status {job_status}.', DEPLOYMENT_JOB_NOT_DEPLOYED_FAIL_TYPE,
+                                       deployment['name'], -1)
 
     def postpone_api_timeout(self, name: str, timeout_timestamp: datetime):
         ultimate_timeout = timeout_timestamp + timedelta(seconds=CORTEX_DELETE_TIMEOUT_SEC)
@@ -398,10 +399,10 @@ class CortexClient:
                             logger.info(print_line)
 
         if job_id is None:
-            cmd_arr = [CORTEX_PATH, "logs", name, f"--env={self.cortex_env}"]
+            cmd_arr = [CORTEX_PATH, "logs", "--yes", name, f"--env={self.cortex_env}"]
 
         else:
-            cmd_arr = [CORTEX_PATH, "logs", name, job_id, f"--env={self.cortex_env}"]
+            cmd_arr = [CORTEX_PATH, "logs", "--yes", name, job_id, f"--env={self.cortex_env}"]
 
         worker = Thread(target=listen_on_logs, args=(cmd_arr,), daemon=True, name=f'api_{name}')
         worker.start()
@@ -532,10 +533,22 @@ class CortexClient:
 @contextmanager
 def open_pg_cursor(db_connection_pool, key=None):
     try:
-        with db_connection_pool.getconn(key) as conn:
-            conn.autocommit = True
-            with conn.cursor() as cur:
-                yield cur
+        for retry in range(DB_TRIES):
+            try:
+                with db_connection_pool.getconn(key) as conn:
+                    conn.autocommit = True
+                    with conn.cursor() as cur:
+                        yield cur
+
+                    break
+
+            except PoolError as e:
+                if retry + 1 >= DB_TRIES:
+                    raise ValueError(f'Retries exhausted with retry {retry}.') from e
+
+                logger.info(f'DB pool error: {e}. Sleeping for an retry {retry}.')
+                # sleep release GIL, allowing other threads to continue
+                time.sleep(DB_RETRY_SEC)
 
     finally:
         db_connection_pool.putconn(conn, key)
